# FILE: link.tcl
#
# An object which implements an event-driven, asynchronous, line-based
# network connection.  It generates an event Readable, when a line
# is available for reading, and queues lines for eventual writing.
# This is the object which actually owns and manages a Tcl socket.
#
# The Tcl event loop must be running for this event-driven object to function.
#

Oc_Class Net_Link {

    # A maximum backlog on the queue, for detecting possible bugs
    common maxQueueLength 100

    # The socket through which we communicate to the host.
    const public variable socket

    # The name of the remote host
    public variable hostname 

    # The remote port
    public variable port 
    
    # line read from socket
    private variable lineFromSocket

    # queue of messages to write to socket
    private variable putQueue = {}

    # Constructor is a non-blocking operation.  If it doesn't return
    # an error, it will generate a Ready event if and when it has established
    # a socket connection.  Then, it will generate Readable events whenever
    # it has a line to be read (using the Get method).  It will generate a
    # Delete event whenever it is destroyed.
    Constructor { args } {
        eval $this Configure $args
        # Try to open connection to host
        if {![info exists socket]} {
            if {![info exists hostname] || ![info exists port]} {
                error "Missing required options: -hostname and -port"
            }
            #
	    # Asynchronous connection of client sockets was broken in
	    # Tcl for Windows until release 8.0.5.  Use -async
	    # connection if it is available.
	    global tcl_platform
	    if {[string match windows $tcl_platform(platform)]
		    && [package vcompare [package provide Tcl] 8.0] <= 0} {

# On Tcl for Alpha Windows NT platforms, [socket -async] appears to still
# be broken in release 8.0.5.  Rather than get in the business of tracking
# machine dependencies, we'll assume use of [socket -async] needs to wait
# until at least release 8.1 of Tcl.
#		global tcl_patchLevel
#		if {[string compare 8.0.5 $tcl_patchLevel]} {
		    # socket -async is broken.  Do blocking open
		    if {[catch {socket $hostname $port} msg]} {
			set msg "Can't connect to $hostname:$port:\n\t$msg"
			error $msg $msg
		    }
		    set socket $msg
		    set event [after 0 $this ConfigureSocket]
		    Oc_EventHandler New _ $this Delete \
			[list after cancel $event] \
			-groups [list $this-ConfigureSocket $this]
		    return
#		}
	    }
	    if {[catch {socket -async $hostname $port} msg]} {
		set msg "Can't connect to $hostname:$port:\n\t$msg"
		error $msg $msg
	    }
	    set socket $msg
	    fileevent $socket writable [list $this VerifyOpen]
        } else {
	    foreach {_ hostname port} [fconfigure $socket -peername] {break}
	    set event [after 0 $this ConfigureSocket]
	    Oc_EventHandler New _ $this Delete \
		[list after cancel $event] \
		-groups [list $this-ConfigureSocket $this]
	}
    }

    # An asynchoronous client socket connection reports that it
    # is open.  Verify that, and if true, configure the socket.
    method VerifyOpen {} {
	global tcl_platform
	fileevent $socket writable {}
	if {![catch {fconfigure $socket -error} msg]} {
	    # -error recognized
	    if {![string match {} $msg]} {
		set msg "Can't connect to $hostname:$port:\n\t$msg"
		Oc_Log Log $msg status $class
		$this Delete
		return
	    }
	}
	# Also check that we can actually write...
	if {[catch {puts -nonewline $socket ""} msg]} {
	    # Write failed
	    set msg "Can't write to $hostname:$port:\n\t$msg"
	    Oc_Log Log $msg status $class
	    $this Delete
	    return
	}

	# Experience shows the only reliable way to be sure we have
	# a connected socket is to try to retrieve information about
	# the peer.  That blocks (boo!) but we need to do it anyway.
	if {[catch {
		foreach {_ hostname port} [fconfigure $socket -peername] {break}
		} msg]} {
	    set msg "Not connected to $hostname:$port\n\t:$msg"
	    Oc_Log Log $msg status $class
	    $this Delete
	    return
	}
	$this ConfigureSocket
    }

    # configure socket for line-oriented traffic, then report the link
    # is ready
    method ConfigureSocket {} {
        Oc_EventHandler DeleteGroup $this-ConfigureSocket
        fconfigure $socket -blocking 0 -buffering line -translation {auto crlf}
        fileevent $socket readable [list $this Receive]
        Oc_Log Log "Connected to $hostname:$port" status $class
        # Be sure to flush any queued messages before the application exits
        Oc_EventHandler New _ Oc_Main Exit [list $this Flush] -oneshot 1 \
                -groups $this
        Oc_EventHandler Generate $this Ready
    }

    method Pause {} {socket} {
	fileevent $socket readable {}
    }

    method Resume {} {socket} {
	fileevent $socket readable [list $this Receive]
    }

    # handler called when there is data available from the host
    # It invokes other events describing what it receives.
    method Receive {} {
        if {[catch {gets $socket line} readCount]} {
            $this SocketError $readCount
            return
        }
        if {$readCount < 0} {
            if {[fblocked $socket]} {
                # 'gets' call is blocked until a whole line is available
                return
            }
            if {[eof $socket]} {
                $this SocketEOF
                return
            }
        }
        set lineFromSocket [Oc_Url Unescape $line]
        Oc_EventHandler Generate $this Readable
    }

    private method SocketError { msg } {
        Oc_Log Log "socket error in connection $this to\
		$hostname:$port:\n\t$msg" status $class
        $this Delete
        return
    }

    private method SocketEOF {} {
        fileevent $socket readable {}
	catch { puts $socket "" } ;# For some reason, this seems to
	## help close the network connection on the other end.
	## -mjd, 98-10-09
        Oc_Log Log "socket closed by $hostname:$port in $this" status $class
        $this Delete
        return
    }

    method Get {} {
        return $lineFromSocket
    }

    method Put { msg } {
        # Escape the newlines in the string -- a mini version of
        # 'Oc_Url Escape'
        regsub -all % $msg %25 msg
        regsub -all "\n" $msg %0a msg
        lappend putQueue $msg

        # Use event-driven writing of message to socket, if possible.
        # Unfortunately the [fileevent ? writable] command is broken
        # on Windows platforms in pre-8.0 versions of Tcl.  See the
        # file ./fewtest.tcl for more information.
        #
        global tcl_platform
        if {[string match windows $tcl_platform(platform)] \
                && [package vsatisfies [package provide Tcl] 7]} {
            $this Send
        } else {
            fileevent $socket writable [list $this Send]
            if {[llength $putQueue] > $maxQueueLength} {
                # will this message ever be seen?
                Oc_Log Log "Maximum queue length exceeded in link\
			$this:\n\tPossibly a socket error?" warning $class
		$this Flush
#                $this Delete
#                return
            }
        }
    }

    method Send {} {
        # This won't block, but could leave data backed up in an internal
        # Tcl buffer.  That's unlikely since we're waiting for a writable
        # socket.
        puts $socket [lindex $putQueue 0]
        set putQueue [lrange $putQueue 1 end]
        if {![llength $putQueue]} {
            fileevent $socket writable {}
        }
    }

    method Flush {} {
        # Flush the queue
        while {[llength $putQueue]} {
            $this Send
        }
    }

    Destructor {
        Oc_EventHandler Generate $this Delete
        Oc_EventHandler DeleteGroup $this
        if {[info exists socket]} {
            fileevent $socket readable {}
            catch { $this Flush }
            Oc_Log Log "Closing socket $socket in $this" status $class
            close $socket
        }
    }

}
