# FILE: threadWish.tcl
#
package require Tk

Oc_Class Net_ThreadWish {
    const public variable thread

    private variable slave
    private variable breakcmds = {}
    private array variable shared = {}

    Constructor {args} {
	eval $this Configure $args

	# Create slave interpreter for the GUI
	set slave [interp create -- slave-$this]

	# set up for slave destruction
	# NOTE: THIS MUST COME BEFORE [package require Oc]
	# in the slave !!!!!
	interp alias $slave exit {} $this Exit

	# Load Tk, so we can rename [tkwait]
	# Hide the main window until we can load the GUI
	$slave eval {
	    package require Tk
	    package require Oc
	    Oc_CommandLine Parse [list -console]
	    console hide
	    wm withdraw .
	}

	# For each nested event loop, stack a command which will break it
	$slave eval {rename tkwait Tk_tkwait}
	interp alias $slave tkwait {} $this Tkwait
	$slave eval {rename vwait Tcl_vwait}
	interp alias $slave vwait {} $this Vwait

	# Commands for interaction with the "thread"
	interp alias $slave appexit {} $thread Send delete
	interp alias $slave remote {} $this Remote
	interp alias $slave share {} $this Share

	# Send for GUI Script
	set id [$thread Send GetGui]
	Oc_EventHandler New _ $thread Reply$id \
		[list $this ReceiveGui 0 $id] -oneshot 1 \
		-groups [list $this $this-$id]
	Oc_EventHandler New _ $thread Timeout$id \
		[list $this ReceiveGui 1 $id] -oneshot 1 \
		-groups [list $this $this-$id]

	# Set up to get Tell messages from remote app
	Oc_EventHandler New _ $thread Readable [list $this Receive] \
		-groups [list $this]

	# Set slave to identify as the remote app
	$slave eval Oc_Main SetAppName [$thread Cget -alias]
	$slave eval {
	    package require Ow
	    wm title . [Oc_Main GetInstanceName]
	    wm iconname . [wm title .]
	    console title "[wm title .] Console"
	    proc VerifyExit {} {
		set response [Ow_Dialog 1 "Exit [Oc_Main GetInstanceName]?" \
			warning "Are you *sure* you want to exit \
			[Oc_Main GetInstanceName]?" {} 1 Yes No]
		if {$response == 0} {
		    appexit
		    exit
		}
	    }
	    wm protocol . WM_DELETE_WINDOW VerifyExit
	    Ow_SetIcon .
	}
    }

    Destructor {
	Oc_EventHandler Generate $this Delete
        Oc_EventHandler DeleteGroup $this

	# Invoke all nested event loop breaking commands, so we
	# break out of all event loops (eventually...)
	foreach cmd $breakcmds {
	    catch {$slave eval $cmd}
	}

	# Convert all commands into [return],
	# so nothing more happens in $slave
	if {[package vsatisfies [package provide Tcl] 7]} {
	    set cmdList [$slave eval info commands]
	    set saveExp {^(proc|return)$}
	} else {
	    set cmdList [$this Commands]
	    set saveExp {^(::proc|::return)$}
	}
	foreach cmd $cmdList {
	    if {[regexp $saveExp $cmd]} {continue}
	    $slave eval [list proc $cmd args [list return -code return]]
	}
	$slave eval [list proc proc args [list return -code return]]
	after idle [list after 1 [list interp delete $slave]]
    }

    callback method ReceiveGui {code id} {
	Oc_EventHandler DeleteGroup $this-$id
	if {!$code} {
            set reply [$thread Get]
	    set code [lindex $reply 0]
	} else {
	    set reply [list 1 "Timed out waiting to receive GUI"]
	}
        if {!$code} { ;# Check for error
	    set script {
	    	Oc_EventHandler New _ Oc_Main Exit \
			{Oc_Log Log {[exit] called from startup script} info} \
			-groups [list startup] -oneshot 1
	    }
	    append script [lindex $reply 1]
	    set code [catch {$slave eval $script} message]
	    # Script evaluation might have destroyed us
	    if {![info exists slave]} {return}
        } else {
	    regexp {[0-9]+} [$thread Cget -pid] id
	    set message "Error retrieving GUI from [$thread Cget \
		    -alias]<$id>:\n[lindex $reply 1]"
	}
	$slave eval {Oc_EventHandler DeleteGroup startup}
	if {$code} {
	    $slave eval [list catch [list Oc_Log Log $message error]]
	}
	if {[info exists slave]} {
	    $slave eval {wm deiconify .; raise .}
	}
    }

    private method Commands {{ns {}}} {
	set cmds [$slave eval [list info commands ${ns}::*]]
	foreach child [$slave eval [list namespace children $ns]] {
	    set cmds [concat $cmds [$this Commands $child]]
	}
	return $cmds
    }

    callback method Exit {args} {
	$thread Configure -gui 0
    }

    callback method Tkwait {sub name} {
	switch -glob -- $sub {
	    va* {
		set cmd [list set $name 0]
	    }
	    vi* {
		set cmd [list event generate $name <Visibility>]
	    }
	    w* {
		set cmd [list destroy $name]
	    }
	}
	set breakcmds [linsert $breakcmds 0 $cmd]
	set code [catch {$slave eval [list Tk_tkwait $sub $name]} msg]
	if {[info exists breakcmds]} {
	    set breakcmds [lrange $breakcmds 1 end]
	}
	return -code $code $msg
    }

    callback method Vwait {name} {
	set breakcmds [linsert $breakcmds 0 [list set $name 0]]
	set code [catch {$slave eval [list Tcl_vwait $name]} msg]
	if {[info exists breakcmds]} {
	    set breakcmds [lrange $breakcmds 1 end]
	}
	return -code $code $msg
    }

    callback method Remote {switch args} {
	if {[string compare -callback $switch]} {
	    eval [list $thread Send Eval {} $switch] $args
	} else {
	    eval [list $thread Send Eval] $args
	}
    }

    callback method Share {remote {local {}}} {
	# Default: use same local name as in the app
	if {[string match {} $local]} {
	    set local $remote
	}
	set trace "[list remote set $remote] \[[list uplevel #0 set $local]] ;#"

	# Set up traces in server
	$thread Send Share $remote $local $trace

	# Set up trace here.
	# Hmmm, this seemed like the right thing to do, but
	# the second time the interface was opened, there was
	# some kind of infinite write trace loop.  So, don't
	# set up a local write trace.  Instead wait for a command
	# from the server to do it for you.  The down side is that
	# any writes to $local before that write trace instruction
	# arrives from the server will not be sent out.
#	$slave eval [list trace variable $local w $trace]

# NOT YET ACTIVE:
if 0 {
	if {[info exists shared($local)]} {
	    # $local is already a shared variable
	    if {[string compare $remote $shared($local)] == 0} {
		# $local is already shadowing $remote.  Nothing to do.
		return
	    }
	    # $local shadows a variable other than $remote.  Stop it.
	    trace vdelete $local w 
	}
}
    }

    method Receive {} {
        set message [$thread Get]
	switch [lindex $message 0] {
	    Result {
		if {[catch {$slave eval [lindex $message 1]} result]} {
		    Oc_Log Log $result error
		}
	    }
	}
    }
}
