/* FILE: messages.cc             -*-Mode: c++-*-
 *
 *      Routines which provide message display services.
 *
 * NOTICE: Please see the file ../../LICENSE
 *
 * Last modified on: $Date: 2000/08/25 17:01:25 $
 * Last modified by: $Author: donahue $
 */

#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include "port.h"
#include "imports.h"
#include "messages.h"
#include "oc.h"

/* End includes */     /* Optional directive to pimake */

#if (SYSTEM_TYPE == WINDOWS)
static int
WindowsMessageBox(const char *buf)
{
  MessageBeep(MB_ICONEXCLAMATION);
  // Title?
  MessageBox(NULL, buf, "OOMMF Error",
             MB_ICONSTOP | MB_OK | MB_TASKMODAL | MB_SETFOREGROUND);
  return TCL_OK;
}

int
Oc_WindowsMessageBoxCmd(ClientData, Tcl_Interp *interp, int argc, char **argv)
{
    Tcl_ResetResult(interp);
    if (argc != 2) {
        Tcl_AppendResult(interp, argv[0], " must be called with"
            " 1 argument: messsage", (char *) NULL);
        return TCL_ERROR;
    }
    return WindowsMessageBox(argv[1]);
}
#endif

/*
 * Return msg, truncated to have at most maxLines lines.
 * Don't write past (msg + bufLength).
 */
static char *
TruncateMessage(char* msg, int maxLines, int bufLength)
{
    const char *tag = "(message truncated)";
    char* p = msg;
    char* limit = msg + bufLength - strlen(tag) - 1;
    int numLines = 0;
    while ((p < limit) && (numLines < maxLines) && (*p != '\0')) {
        if (*p == '\n') {
            numLines++;
        }
        p++;
    }
    if (*p != '\0') {
        strcpy(p, tag);
    }
    return msg;
}

char* panicHeader = NULL;

static const char *
GetPanicHeader()
{
    return panicHeader;
}

void
Oc_SetPanicHeader(const char* header)
{
    delete [] panicHeader;
    panicHeader = new char[strlen(header)+1];
    strcpy(panicHeader, header);
}

int
Oc_SetPanicHeaderCmd(ClientData, Tcl_Interp *interp, int argc, char **argv)
{
    Tcl_ResetResult(interp);
    if (argc != 2) {
        Tcl_AppendResult(interp, argv[0], " must be called with"
            " 1 argument: messsage", (char *) NULL);
        return TCL_ERROR;
    }
    Oc_SetPanicHeader(argv[1]);
    return TCL_OK;
}

/*
 * The routines snprintf() and vsnprintf() supplied by some C libraries
 * are very helpful to prevent buffer overflows.  Here we provide a
 * poor man's substitute which doesn't prevent problems, but detects
 * them and causes a fatal error.  Too bad the real routines aren't ANSI.
 */
int
Oc_Vsnprintf(char *str, size_t n, const char *format, va_list ap)
{
  str[n-1] = '\0';
  int len = OC_SPRINTF_WRAP(vsprintf(str,format,ap));
  if (str[n-1] != '\0') {
    panic(Oc_AutoBuf("Buffer overflow in Oc_Vsnprintf"));
  }
  return len;
}

int
Oc_Snprintf(char *str, size_t n, const char *format, ...)
{
  va_list arg_ptr;
  va_start(arg_ptr,format);
  int len = Oc_Vsnprintf(str,n,format,arg_ptr);
  va_end(arg_ptr);
  return len;
}

static void
CustomPanic(char *format,...)
{
  char buf[4096];	// Good size? -- must be at least long enough for
			// panicHeader plus the buffer overflow message.
			// Really long error stacks might still overflow
			// this.  Consider printing truncated message
			// even if buffer overflows.
  va_list arg_ptr;
  va_start(arg_ptr,format);

  strncpy(buf,GetPanicHeader(),sizeof(buf));
  Oc_Vsnprintf(buf + strlen(buf), sizeof(buf) - strlen(buf), format, arg_ptr);
  va_end(arg_ptr);

#if (SYSTEM_TYPE==WINDOWS)
    WindowsMessageBox(TruncateMessage(buf,20,sizeof(buf)));
    ExitProcess(1);
#else
    fprintf(stderr, TruncateMessage(buf,20,sizeof(buf)));
    fprintf(stderr, "\n");
    fflush(stderr);
    abort();
#endif
}

void
Oc_InitPanic(const char *nameOfExecutable)
{
    char buf[256];
    Oc_Snprintf(buf, sizeof(buf), "<%lu> %s %s panic:\n",
	    (unsigned long) getpid(), nameOfExecutable, OC_VERSION);
    Oc_SetPanicHeader(buf);
    Tcl_SetPanicProc((Tcl_PanicProc *)CustomPanic);
}
